# ------- #
# Imports #
# ------- #

from ordered_set import OrderedSet
from .fns import forEach, raise_


# ---- #
# Main #
# ---- #


class ApSet:
    def __init__(self, aList=[]):
        validateInput(aList)

        #
        # _orderedElements holds a set of tuples
        #   (hashable: boolean, idOrElement) where idOrElement is the id of the
        #   element when hashable is false.  Likewise when hashable is true,
        #   idOrElement is the actual element.
        #
        self._orderedElements = OrderedSet()

        self._hashableElements = set()

        #
        # because this holds non-hashable elements, it needs to be a dict of
        #   id(element) -> element
        #
        self._nonHashableElements = {}
        forEach(self._addElement)(aList)

    def _addElement(self, element):
        try:
            self._hashableElements.add(element)
            self._orderedElements.add((True, element))
        except:
            self._nonHashableElements[id(element)] = element
            self._orderedElements.add((False, id(element)))

        return self

    def __contains__(self, element):
        return self.has(element)

    def __iter__(self):
        return ApSetIterator(self)

    def __len__(self):
        return len(self._orderedElements)

    def add(self, element):
        return self._addElement(element)

    def clear(self):
        self._hashableElements.clear()
        self._nonHashableElements.clear()
        self._orderedElements.clear()
        return self

    def has(self, element):
        try:
            return element in self._hashableElements
        except:
            return id(element) in self._nonHashableElements

    def remove(self, element):
        elementRemoved = False
        try:
            if element in self._hashableElements:
                elementRemoved = True
                self._orderedElements.discard((True, element))
                self._hashableElements.remove(element)
        except:
            if id(element) in self._nonHashableElements:
                elementRemoved = True
                self._orderedElements.discard((False, id(element)))
                del self._nonHashableElements[id(element)]

        if not elementRemoved:
            raise KeyError(
                f"the element '{str(element)}' does not exist in ApSet"
            )

        return self


class ApSetIterator:
    def __init__(self, apSetInst):
        self._inst = apSetInst
        self._orderedElementsIterator = iter(apSetInst._orderedElements)

    def __iter__(self):
        return self

    def __next__(self):
        hashable, elementOrId = next(self._orderedElementsIterator)

        if hashable:
            return elementOrId
        else:
            return self._inst._nonHashableElements[elementOrId]


# ------- #
# Helpers #
# ------- #


def validateInput(aList):
    if not isinstance(aList, list):
        raise_(
            ValueError,
            f"""
            aList is not an instance of list
            type given: {type(aList)}
            """,
        )



from inspect import signature
from ..internal.raise_ import raise_
import wrapt


def argIsType(aType):
    @wrapt.decorator
    def wrapper(fn, _instance, args, kwargs):
        typePassed = type(args[0])
        if typePassed is not str:
            argName = list(signature(fn).parameters)[0]
            fnName = fn.__name__
            typeName = aType.__name__
            raise_(
                ValueError,
                f"""\
                {fnName} requires {argName} to have the type {typeName}
                type passed: {typePassed.__name__}
                """,
            )

        return fn(*args, **kwargs)

    return wrapper



from .forEach import forEach
from .isLaden import isLaden
from .joinWith import joinWith
from .map_ import map_
from .passThrough import passThrough
from .raise_ import raise_

__all__ = ["forEach", "isLaden", "joinWith", "map_", "passThrough", "raise_"]



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['all_purpose_set',
 'all_purpose_set.fns',
 'all_purpose_set.fns.decorators',
 'all_purpose_set.fns.internal']

package_data = \
{'': ['*']}

install_requires = \
['ordered_set>=3.1,<4.0', 'tedent>=0.1.1,<0.2.0', 'wrapt>=1.10,<2.0']

setup_kwargs = {
    'name': 'all-purpose-set',
    'version': '0.1.5',
    'description': 'A set that works with both hashable and non-hashable elements',
    'long_description': '# All Purpose Set\n\n*Note: This document is best viewed [on github](https://github.com/olsonpm/py_all-purpose-set).\nPypi\'s headers are all caps which presents inaccurate information*\n\n\n<br>\n\n<!-- START doctoc generated TOC please keep comment here to allow auto update -->\n<!-- DON\'T EDIT THIS SECTION, INSTEAD RE-RUN doctoc TO UPDATE -->\n**Table of Contents**\n\n- [What is it?](#what-is-it)\n- [Why create it?](#why-create-it)\n- [Simple usage](#simple-usage)\n- [See also](#see-also)\n- [Api](#api)\n- [Test](#test)\n\n<!-- END doctoc generated TOC please keep comment here to allow auto update -->\n\n<br>\n\n### What is it?\n\n- A set which doesn\'t require hashable contents\n\n<br>\n\n### Why create it?\n\n- I often have a need to store non-hashable contents in a set.  For example\n  storing a dict isn\'t possible with the builtin set.\n\n  ```py\n  # doesn\'t work\n  someDict = { "key": "value" }\n  someSet = { someDict }\n  ```\n\n<br>\n\n### Simple usage\n\n```py\nfrom all_purpose_set import ApSet\n\nsomeDict = { "key": "value" }\nsomeSet = ApSet([someDict])\n\nprint(someDict in someSet) # prints True\n```\n\n<br>\n\n### See also\n\n- [All Purpose Dict](https://github.com/olsonpm/py_all-purpose-dict)\n\n<br>\n\n### Api\n\n*Note: This api is young and subject to change quite a bit.  There also may be\nfunctionality present in the builtin set which this set doesn\'t cover.  I\'m\nwilling to add it so please just raise a github issue or PR with details.*\n\n#### class ApSet([a list])\n- all methods return `self` unless specified otherwise\n- iterates in the order of insertion\n- currently the internal methods implemented are\n  - \\_\\_contains\\_\\_\n  - \\_\\_iter\\_\\_\n  - \\_\\_len\\_\\_\n\n##### add(something)\n\n##### clear()\n\n##### has(something) => bool\n- a function alternative to `key in aSet`\n\n##### remove(something)\n- raises a `KeyError` if the element doesn\'t exist\n\n<br>\n\n### Test\n\n```sh\n#\n# you must have poetry installed\n#\n$ poetry shell\n$ poetry install\n$ python runTests.py\n```\n',
    'author': 'Philip Olson',
    'author_email': 'philip.olson@pm.me',
    'url': 'https://github.com/olsonpm/py_all-purpose-set',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
